@@ -1,9 +1,9 @@
 #' spatially or temporally aggregate stars object
 #' 
-#' spatially or temporally aggregate stars object, returning a data cube with lower spatial or temporal resolution 
+#' spatially or temporally aggregate stars object, returning a data cube with lower spatial or temporal resolution
 #' 
 #' @param x object of class \code{stars} with information to be aggregated
-#' @param by object of class \code{sf} or \code{sfc} for spatial aggregation, for temporal aggregation a vector with time values (\code{Date}, \code{POSIXct}, or \code{PCICt}) that is interpreted as a sequence of left-closed, right-open time intervals or a string like "months", "5 days" or the like (see \link{cut.POSIXt}); if by is an object of class \code{stars}, it is converted to sfc by \code{st_as_sfc(by, as_points = FALSE)} thus ignoring its time component.
+#' @param by object of class \code{sf} or \code{sfc} for spatial aggregation, for temporal aggregation a vector with time values (\code{Date}, \code{POSIXct}, or \code{PCICt}) that is interpreted as a sequence of left-closed, right-open time intervals or a string like "months", "5 days" or the like (see \link{cut.POSIXt}); if by is an object of class \code{stars}, it is converted to sfc by \code{st_as_sfc(by, as_points = FALSE)} thus ignoring its time component. Note: each pixel is assigned to only a single group (in the order the groups occur) so non-overlapping spatial features and temporal windows are recommended.
 #' @param FUN aggregation function, such as \code{mean}
 #' @param ... arguments passed on to \code{FUN}, such as \code{na.rm=TRUE}
 #' @param drop logical; ignored
@@ -26,7 +26,8 @@
 #' by_t = "2 days"
 #' x_agg_time2 = aggregate(x, by = by_t, FUN = max) 
 #' st_get_dimension_values(x_agg_time2, "time")
-#' x_agg_time - x_agg_time2
+#' #TBD:
+#' #x_agg_time - x_agg_time2
 #'
 #' # aggregate time dimension in format POSIXct
 #' x = st_set_dimensions(x, 4, values = as.POSIXct(c("2018-07-31", 
@@ -37,32 +38,52 @@
 #' by_t = as.POSIXct(c("2018-07-31", "2018-08-02"))
 #' x_agg_posix = aggregate(x, by = by_t, FUN = max)
 #' st_get_dimension_values(x_agg_posix, "time")
-#' x_agg_time - x_agg_posix
+#' #TBD:
+#' # x_agg_time - x_agg_posix
 #' aggregate(x, "2 days", mean)
-#' # Spatial aggregation, see https://github.com/r-spatial/stars/issues/299
-#' prec_file = system.file("nc/test_stageiv_xyt.nc", package = "stars")
-#' prec = read_ncdf(prec_file, curvilinear = c("lon", "lat"))
-#' prec_slice = dplyr::slice(prec, index = 17, along = "time")
-#' nc = sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"), "nc.gpkg")
-#' nc = st_transform(nc, st_crs(prec_slice))
-#' agg = aggregate(prec_slice, st_geometry(nc), mean)
+#' if (require(ncmeta, quietly = TRUE)) {
+#'  # Spatial aggregation, see https://github.com/r-spatial/stars/issues/299
+#'  prec_file = system.file("nc/test_stageiv_xyt.nc", package = "stars")
+#'  prec = read_ncdf(prec_file, curvilinear = c("lon", "lat"))
+#'  prec_slice = dplyr::slice(prec, index = 17, along = "time")
+#'  nc = sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"), "nc.gpkg")
+#'  nc = st_transform(nc, st_crs(prec_slice))
+#'  agg = aggregate(prec_slice, st_geometry(nc), mean)
+#'  plot(agg)
+#' }
+#'
+#' # example of using a function for "by": aggregate by month-of-year
+#' d = c(10, 10, 150)
+#' a = array(rnorm(prod(d)), d) # pure noise
+#' times = Sys.Date() + seq(1, 2000, length.out = d[3])
+#' m = as.numeric(format(times, "%m"))
+#' signal = rep(sin(m / 12 * pi), each = prod(d[1:2])) # yearly period
+#' s = (st_as_stars(a) + signal) %>%
+#'       st_set_dimensions(3, values = times)
+#' f = function(x, format = "%B") {
+#' 	  months = format(as.Date(paste0("01-", 1:12, "-1970")), format)
+#' 	  factor(format(x, format), levels = months)
+#' }
+#' agg = aggregate(s, f, mean)
 #' plot(agg)
 aggregate.stars = function(x, by, FUN, ..., drop = FALSE, join = st_intersects, 
 		as_points = any(st_dimension(by) == 2, na.rm = TRUE), rightmost.closed = FALSE,
 		left.open = FALSE, exact = FALSE) {
 
-	if (inherits(by, "stars")) {
-		by = st_as_sfc(by, as_points = FALSE)
-		# and if not, then use st_normalize(by)
-	}
-
-	classes = c("sf", "sfc", "POSIXct", "Date", "PCICt", "character")
-	if (!inherits(by, classes))
+	classes = c("sf", "sfc", "POSIXct", "Date", "PCICt", "character", "function", "stars")
+	if (!is.function(by) && !inherits(by, classes))
 		stop(paste("currently, only `by' arguments of class", 
 			paste(classes, collapse= ", "), "supported"))
+	if (inherits(by, "stars"))
+		by = st_as_sfc(by, as_points = FALSE) # and if not, then use st_normalize(by)
+
+	if (inherits(by, "sf")) {
+		geom = attr(by, "sf_column")
+		by = st_geometry(by)
+	} else
+		geom = "geometry"
+	stopifnot(!missing(FUN), is.function(FUN))
 
-	if (missing(FUN))
-		stop("missing FUN argument")
 	if (exact && inherits(by, c("sf", "sfc_POLYGON", "sfc_MULTIPOLYGON")) && has_raster(x)) {
     	if (!requireNamespace("raster", quietly = TRUE))
         	stop("package raster required, please install it first") # nocov
@@ -74,22 +95,29 @@ aggregate.stars = function(x, by, FUN, ..., drop = FALSE, join = st_intersects,
 		e = exactextractr::coverage_fraction(as(r, "Raster"), by)
 		st = do.call(raster::stack, e)
 		m = raster::getValues(st)
-		if (identical(FUN, mean))
-			m = sweep(m, 2, colSums(m), "/") # mean: divide weights by the sum of weights
-		else if (!identical(FUN, sum))
-			stop("for exact=TRUE, FUN should either be mean or sum")
+		#if (identical(FUN, mean)) see https://github.com/r-spatial/stars/issues/289
+		if (!identical(FUN, sum)) {
+			if (isTRUE(as.character(as.list(FUN)[[3]])[2] == "mean"))
+				m = sweep(m, 2, colSums(m), "/") # mean: divide weights by the sum of weights
+			else
+				stop("for exact=TRUE, FUN should either be mean or sum")
+		}
 		new_dim = c(prod(dim(x)[1:2]), prod(dim(x)[-(1:2)]))
 		out_dim = c(ncol(m), dim(x)[-(1:2)])
 		if (isTRUE(list(...)$na.rm))
 			x = lapply(x, function(y) { y[is.na(y)] = 0.0; y })
 		agg = lapply(x, function(a) array(t(m) %*% array(a, dim = new_dim), dim = out_dim))
-		ret = st_as_stars(agg, dimensions = 
-			create_dimensions(append(list(sfc = create_dimension(values = by)),
-			st_dimensions(x)[-(1:2)])))
-		return(ret)
+		# %*% dropped units, so to propagate units, if present we need to copy (mean/sum):
+		d = create_dimensions(append(setNames(list(create_dimension(values = by)), geom),
+			st_dimensions(x)[-(1:2)]))
+		for (i in seq_along(x)) {
+			if (inherits(x[[i]], "units")) 
+				agg[[i]] = units::set_units(agg[[i]], units(x[[i]]), mode = "standard")
+			names(dim(agg[[i]])) = names(d)
+		}
+		return(st_as_stars(agg, dimensions = d))
 	}
 
-	geom = "geometry"
 	drop_y = FALSE
 	grps = if (inherits(by, c("sf", "sfc"))) {
 			x = if (has_raster(x)) {
@@ -100,11 +128,6 @@ aggregate.stars = function(x, by, FUN, ..., drop = FALSE, join = st_intersects,
 					ndims = 1
 					st_upfront(x, which_sfc(x))
 				}
-
-			if (inherits(by, "sf")) {
-				geom = attr(by, "sf_column")
-				by = st_geometry(by)
-			}
 	
 			# find groups:
 			# don't use unlist(join(x_geoms, by)) as this would miss the empty groups, 
@@ -123,13 +146,17 @@ aggregate.stars = function(x, by, FUN, ..., drop = FALSE, join = st_intersects,
 			ndims = 1
 			x = st_upfront(x, which_time(x))
 			values = expand_dimensions(x)[[1]]
-			if (inherits(by, "character")) {
+			if (inherits(by, "function")) {
+				i = by(values)
+				if (!is.factor(i))
+					i = as.factor(i)
+				by = levels(i)
+			} else if (inherits(by, "character")) {
 				i = cut(values, by, right = left.open)
 				by = if (inherits(values, "Date"))
 						as.Date(levels(i))
 					else
 						as.POSIXct(levels(i))
-				i = as.integer(i)
 			} else {
 				if (!inherits(values, class(by)))
 					warning(paste0('argument "by" is of a different class (', class(by)[1], 
@@ -137,7 +164,7 @@ aggregate.stars = function(x, by, FUN, ..., drop = FALSE, join = st_intersects,
 				i = findInterval(values, by, left.open = left.open, rightmost.closed = rightmost.closed)
 				i[ i == 0 | i == length(by) ] = NA
 			}
-			i
+			as.integer(i)
 		}
 
 	d = st_dimensions(x)
@@ -157,20 +184,33 @@ aggregate.stars = function(x, by, FUN, ..., drop = FALSE, join = st_intersects,
 	# rearrange:
 	x = structure(x, dimensions = NULL, class = NULL) # unclass
 	newdims = c(prod(dims[1:ndims]), prod(dims[-(1:ndims)]))
-	for (i in seq_along(x))
-		x[[i]] = agr_grps(array(x[[i]], newdims), grps, seq_along(by), FUN, ...)
+	for (i in seq_along(x)) {
+		a = array(x[[i]], newdims)
+		u = if (inherits(x[[i]], "units") && dim(a)[2] > 0) {
+				a = units::set_units(a, units(x[[i]]), mode = "standard")
+				try(out <- FUN(a[,1], ...))
+				if (inherits(out, "units"))
+					units(out)
+				else
+					NULL
+			} else
+				NULL
+		x[[i]] = agr_grps(a, grps, seq_along(by), FUN, ...)
+		if (is.numeric(x[[i]]) && !is.null(u))
+			x[[i]] = units::set_units(x[[i]], u, mode = "standard")
+	}
 
 	# reconstruct dimensions table:
 	d[[1]] = create_dimension(values = by)
-	names(d)[1] = if (inherits(by, c("POSIXct", "Date", "PCICt")))
+	names(d)[1] = if (inherits(by, c("POSIXct", "Date", "PCICt", "function")))
 			"time"
 		else
 			geom
 	if (drop_y)
 		d = d[-2] # y
 
-	newdim = c(sfc = length(by), dims[-(1:ndims)])
-	st_stars(lapply(x, array, dim = newdim), dimensions = d)
+	newdim = setNames(c(sfc = length(by), dims[-(1:ndims)]), names(d))
+	st_stars(lapply(x, structure, dim = newdim), dimensions = d)
 }
 
 	# aggregate is done over one or more dimensions
@@ -191,13 +231,20 @@ aggregate.stars = function(x, by, FUN, ..., drop = FALSE, join = st_intersects,
 
 #' @export
 aggregate.stars_proxy = function(x, by, FUN, ...) {
-	if (inherits(by, "stars"))
-		by = st_as_sfc(by, as_points = FALSE)
-	if (!inherits(by, c("sf", "sfc", "sfg")))
-		stop("aggregate.stars_proxy only implemented for spatial `by' arguments")
-	by = st_geometry(by)
+	if (!inherits(by, c("sf", "sfc", "sfg", "stars")))
+		collect(x, match.call(), "aggregate", c("x", "by", "FUN"), env = environment(), ...)
+	else {
+		if (inherits(by, "stars"))
+			by = st_as_sfc(by, as_points = FALSE)
+		by = st_geometry(by)
 
-	# this assumes the result is small, no need to proxy
-	l = lapply(seq_along(by), function(i) aggregate(st_as_stars(x[by[i]]), by[i], FUN, ...))
-	do.call(c, c(l, along = list(which_sfc(l[[1]]))))
+		# this assumes each result of a [ selection is small enough to hold in memory
+		l = lapply(seq_along(by), 
+		   	function(i) {
+			   	sel_i = st_normalize(st_as_stars(x[by[i]]))
+			   	aggregate(sel_i, by[i], FUN, ...)
+		   	}
+		)
+		do.call(c, c(l, along = list(which_sfc(l[[1]]))))
+	}
 }
