@@ -1,17 +1,51 @@
 maybe_normalizePath = function(.x, np = FALSE) {
-	prefixes = c("NETCDF:", "HDF5:", "HDF4:", "HDF4_EOS:", "SENTINEL2_L1", "SENTINEL2_L2", "GPKG:")
+	prefixes = c("NETCDF:", "HDF5:", "HDF4:", "HDF4_EOS:", "SENTINEL2_L1", "SENTINEL2_L2", "GPKG:", "/vsi", "http://", "https://")
 	has_prefix = function(pf, x) substr(x, 1, nchar(pf)) == pf
-	if (!np || any(sapply(prefixes, has_prefix, x = .x)))
+	if (is.function(.x) || !np || any(sapply(prefixes, has_prefix, x = .x)))
 		.x
 	else
-		normalizePath(.x, mustWork = FALSE)
+		setNames(normalizePath(.x, mustWork = FALSE), names(.x))
+}
+
+enc2utf8char = function(x) {
+	if (is.character(x))
+		enc2utf8(x)
+	else
+		x
+}
+
+get_names = function(x) {
+	if (is.function(x))
+		x()
+	else
+		x
+}
+
+is_functions = function(x) {
+	is.function(x) || all(sapply(x, is.function))
+}
+
+# strip common start & end parts of x, if remainder != ""
+unique_part = function(x, prefix) {
+	stopifnot(is.character(x), length(x) > 1)
+	original = x
+	while (nchar(x[1]) && length(unique(substr(x, 1, 1))) == 1)
+		x = sapply(x, substring, 2)
+	while ((n <- nchar(x[1])) && length(unique(substr(x, n, n))) == 1)
+		x = sapply(x, substring, 1, n - 1)
+	if (any(x == ""))
+		original
+	else if (is.character(prefix))
+		paste0(prefix, x)
+	else
+		x
 }
 
 
 #' read raster/array dataset from file or connection
 #'
 #' read raster/array dataset from file or connection
-#' @param .x character vector with name(s) of file(s) or data source(s) to be read
+#' @param .x character vector with name(s) of file(s) or data source(s) to be read, or a function that returns such a vector
 #' @param options character; opening options
 #' @param driver character; driver to use for opening file. To override fixing for subdatasets and autodetect them as well, use \code{NULL}.
 #' @param sub character, integer or logical; name, index or indicator of sub-dataset(s) to be read
@@ -26,19 +60,24 @@ maybe_normalizePath = function(.x, np = FALSE) {
 #' @param curvilinear length two character vector with names of subdatasets holding longitude and latitude values for all raster cells, or named length 2 list holding longitude and latitude matrices; the names of this list should correspond to raster dimensions referred to
 #' @param normalize_path logical; if \code{FALSE}, suppress a call to \link{normalizePath} on \code{.x}
 #' @param RAT character; raster attribute table column name to use as factor levels
+#' @param tolerance numeric; passed on to \link{all.equal} for comparing dimension parameters.
 #' @param ... passed on to \link{st_as_stars} if \code{curvilinear} was set
+#' @param exclude character; vector with category value(s) to exclude
+#' @param shorten logical or character; if \code{TRUE} and \code{length(.x) > 1}, remove common start and end parts of array names; if character a new prefix
 #' @return object of class \code{stars}
 #' @details In case \code{.x} contains multiple files, they will all be read and combined with \link{c.stars}. Along which dimension, or how should objects be merged? If \code{along} is set to \code{NA} it will merge arrays as new attributes if all objects have identical dimensions, or else try to merge along time if a dimension called \code{time} indicates different time stamps. A single name (or positive value) for \code{along} will merge along that dimension, or create a new one if it does not already exist. If the arrays should be arranged along one of more dimensions with values (e.g. time stamps), a named list can passed to \code{along} to specify them; see example.
 #'
 #' \code{RasterIO} is a list with zero or more of the following named arguments:
 #' \code{nXOff}, \code{nYOff} (both 1-based: the first row/col has offset value 1),
-#' \code{nXSize}, \code{nYSize}, \code{nBufXSize}, \code{nBufYSize}, \code{bands}, code{resample}.
-#' see https://www.gdal.org/classGDALDataset.html#a80d005ed10aefafa8a55dc539c2f69da for their meaning;
-#' \code{bands} is an integer vector containing the band numbers to be read (1-based: first band is 1)
+#' \code{nXSize}, \code{nYSize}, \code{nBufXSize}, \code{nBufYSize}, \code{bands}, \code{resample}.
+#' See \url{https://gdal.org/doxygen/classGDALDataset.html} for their meaning;
+#' \code{bands} is an integer vector containing the band numbers to be read (1-based: first band is 1).
 #' Note that if \code{nBufXSize} or \code{nBufYSize} are specified for downsampling an image,
 #' resulting in an adjusted geotransform. \code{resample} reflects the resampling method and
 #' has to be one of: "nearest_neighbour" (the default),
 #' "bilinear", "cubic", "cubic_spline", "lanczos", "average", "mode", or "Gauss".
+#'
+#' Data that are read into memory (\code{proxy=FALSE}) are read into a numeric (double) array, except for categorical variables which are read into an numeric (integer) array of class \code{factor}.
 #' @export
 #' @examples
 #' tif = system.file("tif/L7_ETMs.tif", package = "stars")
@@ -75,16 +114,42 @@ maybe_normalizePath = function(.x, np = FALSE) {
 #' file.remove(tmp)
 read_stars = function(.x, ..., options = character(0), driver = character(0),
 		sub = TRUE, quiet = FALSE, NA_value = NA_real_, along = NA_integer_,
-		RasterIO = list(), proxy = !length(curvilinear) && is_big(.x, sub = sub, driver=driver, ...),
-		curvilinear = character(0), normalize_path = TRUE, RAT = character(0)) {
+		RasterIO = list(), proxy = is_functions(.x) || (!length(curvilinear) &&
+				is_big(.x, sub = sub, driver = driver, normalize_path = normalize_path, ...)),
+		curvilinear = character(0), normalize_path = TRUE, RAT = character(0),
+		tolerance = 1e-10, exclude = "", shorten = TRUE) {
+
+	is_false = function(x) { length(x) == 1 && is.logical(x) && !x }
+	if (!is.list(.x) && length(.x) > 1 && is.null(names(.x)) && !is_false(shorten) &&
+				any((short_names <- unique_part(.x, shorten)) != .x))
+		names(.x) = short_names
 
 	x = if (is.list(.x)) {
-			f = function(y, np) enc2utf8(maybe_normalizePath(y, np))
-			rapply(.x, f, classes = "character", how = "replace", np = normalize_path)
+			f = function(y, np) enc2utf8char(maybe_normalizePath(y, np))
+			rapply(.x, f, classes = c("character", "function"), how = "replace", np = normalize_path)
 		} else
-			enc2utf8(maybe_normalizePath(.x, np = normalize_path))
+			enc2utf8char(maybe_normalizePath(.x, np = normalize_path))
 
-	if (length(curvilinear) == 2 && is.character(curvilinear)) {
+	if (missing(curvilinear)) { # see if we can get it from metadata item "GEOLOCATION":
+		geolocation <- try(gdal_metadata(.x, "GEOLOCATION"), silent = TRUE)
+		if (!inherits(geolocation, "try-error")) { # the thing has x and y arrays:
+			# read coordinate arrays:
+			gxy = c(geolocation$X_DATASET, geolocation$Y_DATASET)
+			lon = adrop(read_stars(gxy[1], driver = driver, quiet = quiet, NA_value = NA_value,
+				RasterIO = RasterIO, proxy = FALSE, curvilinear = character(0), ...))
+			lat = adrop(read_stars(gxy[2], driver = driver, quiet = quiet, NA_value = NA_value,
+				RasterIO = RasterIO, proxy = FALSE, curvilinear = character(0), ...))
+			if (all(dim(lon) > 1) && all(dim(lat) > 1)) # both are 2-D matrices:
+				curvilinear = setNames(c(st_set_dimensions(lon, names = c("x", "y")),
+					st_set_dimensions(lat, names = c("x", "y"))), c("x", "y"))
+			else {
+				ulon = unique(diff(as.vector(lon[[1]])))
+				ulat = unique(diff(as.vector(lat[[1]])))
+				if (length(ulon) > 1 || length(ulat) > 1)
+					warning("ignoring irregular (rectilinear) x and/or y coordinates!")
+			}
+		}
+	} else if (length(curvilinear) == 2 && is.character(curvilinear)) { # user-set, read them
 		lon = adrop(read_stars(.x, sub = curvilinear[1], driver = driver, quiet = quiet, NA_value = NA_value,
 			RasterIO = RasterIO, proxy = FALSE, ..., sub_required = TRUE))
 		lat = adrop(read_stars(.x, sub = curvilinear[2], driver = driver, quiet = quiet, NA_value = NA_value,
@@ -92,15 +157,19 @@ read_stars = function(.x, ..., options = character(0), driver = character(0),
 		curvilinear = setNames(c(st_set_dimensions(lon, names = c("x", "y")),
 			st_set_dimensions(lat, names = c("x", "y"))), c("x", "y"))
 	}
+	if (length(curvilinear) && proxy)
+		warning("proxy = TRUE may not work for curvilinear rasters")
 
-	if (length(x) > 1) { # loop over data sources:
+	if (length(x) > 1) { # loop over data sources and RETURNS:
 		ret = lapply(x, read_stars, options = options, driver = driver, sub = sub, quiet = quiet,
 			NA_value = NA_value, RasterIO = as.list(RasterIO), proxy = proxy, curvilinear = curvilinear,
-			along = if (length(along) > 1) along[-1] else NA_integer_)
-		return(do.call(c, append(ret, list(along = along))))
+			along = if (length(along) > 1) along[-1] else NA_integer_, normalize_path = normalize_path)
+		return(do.call(c, append(ret, list(along = along, tolerance = tolerance))))
 	}
 
-	data = sf::gdal_read(x, options = options, driver = driver, read_data = !proxy,
+	# else:
+	data = sf::gdal_read(get_names(x),
+		options = options, driver = driver, read_data = !proxy,
 		NA_value = NA_value, RasterIO_parameters = as.list(RasterIO))
 	if (!is.null(data$default_geotransform) && data$default_geotransform == 1) {
 		## we have the 0 1 0 0 0 1 transform indicated
@@ -146,7 +215,7 @@ read_stars = function(.x, ..., options = character(0), driver = character(0),
 			else
 				ret
 		}
-	} else { # we have one single array:
+	} else { # we have one single array: read it
 		if (!isTRUE(sub)) {
 			sub_required = if (is.null(list(...)$sub_required))
 					FALSE
@@ -163,7 +232,7 @@ read_stars = function(.x, ..., options = character(0), driver = character(0),
 			else
 				get_data_units(attr(data, "data")) # extract data array; sets units if present
 		if (meta_data$driver[1] == "netCDF")
-			meta_data = parse_netcdf_meta(meta_data, x) # sets all kind of units
+			meta_data = parse_netcdf_meta(meta_data, get_names(x)) # sets all kind of units
 		if (! proxy && !is.null(meta_data$units) && !is.na(meta_data$units)
 				&& !inherits(data, "units")) # set units
 			units(data) = try_as_units(meta_data$units)
@@ -176,34 +245,58 @@ read_stars = function(.x, ..., options = character(0), driver = character(0),
 		# handle color table and/or attribute table
 		ct = meta_data$color_tables
 		at = meta_data$attribute_tables
+		if (meta_data$driver[1] == "AIG" && identical(names(at[[1]]), c("VALUE", "COUNT")))
+			at = list() # skip it: https://github.com/r-spatial/stars/issues/435
 		# FIXME: how to handle multiple color, category or attribute tables?
 		if (!proxy && (any(lengths(ct) > 0) || any(lengths(at) > 0))) {
-			min_value = if (!is.null(meta_data$ranges) && meta_data$ranges[1,2] == 1)
-					meta_data$ranges[1,1]
-				else
-					min(data, na.rm = TRUE)
-			data[data < min_value] = NA
-			if (min_value < 0)
-				stop("categorical values should have minimum value >= 0")
-			if (any(lengths(ct) > 0)) {
-				ct = ct[[ which(length(ct) > 0)[1] ]]
-				co = apply(ct, 1, function(x) rgb(x[1], x[2], x[3], x[4], maxColorValue = 255))
-				if (min_value > 0)
-					co = co[-seq_len(min_value)] # removes [0,...,(min_value-1)]
-				f = factor(as.vector(data), levels = seq(min_value, length.out = length(co)))
-				data = structure(f, dim = dim(data), colors = co, class = "factor")
+			if (any(meta_data$ranges[1, c(2,4)] == 1)) { # adjust data to min/max values from image metadata
+				r = range(data, na.rm = TRUE)
+				min_value = if (meta_data$ranges[1,2] == 1)
+						meta_data$ranges[1,1]
+					else
+						r[1]
+				max_value = if (meta_data$ranges[1,4] == 1)
+						meta_data$ranges[1,3]
+					else
+						r[2]
+				data[data < min_value | data > max_value] = NA
 			}
+
+			# convert color table ct to a vector of R colors:
+			co = if (any(lengths(ct) > 0)) {
+				ct = ct[[ which(length(ct) > 0)[1] ]]
+				apply(ct, 1, function(x) rgb(x[1], x[2], x[3], x[4], maxColorValue = 255))
+			} else
+				NULL
+
 			if (any(lengths(at) > 0)) {
+				# select attribute table:
 				which.at = which(lengths(at) > 0)[1]
 				which.column = if (length(RAT))
 						RAT
 					else
 						which(sapply(at[[which.at]], class) == "character")[1]
-				at = at[[ which.at ]][[ which.column ]]
-				if (min_value > 0)
-					at = at[-seq_len(min_value)]
-				attr(data, "levels") = at
+				labels = at = at[[ which.at ]][[ which.column ]]
+				levels = 0:(length(at) - 1)
+				if (length(exclude)) {
+					ex = at %in% exclude
+					labels = labels[!ex]
+					levels = levels[!ex]
+					if (!is.null(co))
+						co = co[!ex]
+				} else
+					ex = rep(FALSE, length(levels))
+			} else {
+				levels = sort(unique(data))
+				labels = as.character(levels)
+				ex = rep(FALSE, length(levels))
 			}
+			# f = factor(as.vector(data), levels = levels, labels = labels) # is too costly;
+			# see https://github.com/r-spatial/stars/issues/565:
+			# construct factor array manually:
+			data = structure(match(as.integer(as.vector(data)), levels),
+							 levels = labels, dim = dim(data), 
+							 colors = co, exclude = ex, class = "factor")
 		}
 
 		dims = if (proxy) {
@@ -220,13 +313,20 @@ read_stars = function(.x, ..., options = character(0), driver = character(0),
 				NULL
 
 		# return:
-		ret = if (proxy) # no data present, subclass of "stars":
-			st_stars_proxy(setNames(list(x), names(.x) %||% tail(strsplit(x, '[\\\\/]+')[[1]], 1)),
-				create_dimensions_from_gdal_meta(dims, meta_data), NA_value = NA_value,
-				resolutions = NULL)
-		else
-			st_stars(setNames(list(data), names(.x) %||% tail(strsplit(x, '[\\\\/:]+')[[1]], 1)),
-				create_dimensions_from_gdal_meta(dim(data), meta_data))
+		name_x = if (!is.null(meta_data$long_name) && !is.na(meta_data$long_name))
+				meta_data$long_name
+			else if (is.function(.x))
+				names(.x()) %||% .x()
+			else
+				x
+		name_x = tail(strsplit(name_x, '[\\\\/:]+')[[1]], 1)
+		ret = if (proxy) { # no data present, subclass of "stars":
+				st_stars_proxy(setNames(list(x), names(.x) %||% name_x),
+					create_dimensions_from_gdal_meta(dims, meta_data), NA_value = NA_value,
+					resolutions = NULL, RasterIO = RasterIO, file_dim = dims)
+			} else
+				st_stars(setNames(list(data), names(.x) %||% name_x),
+					create_dimensions_from_gdal_meta(dim(data), meta_data))
 
 		if (is.list(curvilinear))
 			st_as_stars(ret, curvilinear = curvilinear, ...)
@@ -247,13 +347,17 @@ is_big = function(x, ..., sub = sub, n_proxy = options("stars.n_proxy")[[1]] %||
 get_data_units = function(data) {
 	units = unique(attr(data, "units")) # will fail parsing in as_units() when more than one
 	if (length(units) > 1) {
-		warning(paste("more than one unit available for subdataset: using only", units[1])) # nocov
-		units = units[1] # nocov
-	}
-	if (!is.null(units) && nzchar(units))
-		units = try_as_units(units)
-	if (inherits(units, "units"))
-		units::set_units(structure(data, units = NULL), units, mode = "standard")
-	else
+		warning(paste("more than one unit available for array: ignoring all")) # nocov
 		structure(data, units = NULL)
+	} else {
+		if (!is.null(units) && nzchar(units)) {
+			if (units == "arc-second") # datum grids may have this
+				units = "arcsecond"
+			units = try_as_units(units)
+		}
+		if (inherits(units, "units"))
+			units::set_units(structure(data, units = NULL), units, mode = "standard")
+		else
+			structure(data, units = NULL)
+	}
 }
