@@ -96,7 +96,7 @@
 	args[["drop"]] = FALSE
 	for (i in names(x))
 		x[[i]] = structure(eval(rlang::expr(x[[i]][ !!!args ])), levels = attr(x[[i]], "levels"),
-			colors = attr(x[[i]], "colors"))
+			colors = attr(x[[i]], "colors"), exclude = attr(x[[i]], "exclude"))
 
 	# now do dimensions:
 	if (do_select) {
@@ -138,7 +138,7 @@
 	if (inherits(i, "stars_proxy"))
 		i = st_as_stars(i, downsample = downsample)
 	if (!inherits(i, "stars"))
-		stop("selector should be a stars object")
+		stop("selector i should be a stars object")
 	fun = function(x, y, value) { x[y] = value; x }
 	st_as_stars(mapply(fun, x, i, value = value, SIMPLIFY = FALSE), dimensions = st_dimensions(x))
 }
