@@ -1,9 +1,8 @@
 ## ----setup, include=FALSE-----------------------------------------------------
-knitr::opts_chunk$set(echo = TRUE)
-knitr::opts_chunk$set(collapse = TRUE)
+knitr::opts_chunk$set(echo = TRUE, collapse = TRUE, dev = "png")
+ev = suppressWarnings(require(starsdata, quietly = TRUE))
 knitr::opts_chunk$set(fig.height = 4.5)
 knitr::opts_chunk$set(fig.width = 6)
-ev = suppressWarnings(require(starsdata, quietly = TRUE))
 
 ## -----------------------------------------------------------------------------
 library(stars)
@@ -28,9 +27,9 @@ as.tbl_cube(x)
 merge(x.spl)
 
 ## -----------------------------------------------------------------------------
-merge(x.spl) %>% 
-  setNames(names(x)) %>%
-  st_set_dimensions(3, values = paste0("band", 1:6)) %>%
+merge(x.spl) |>
+  setNames(names(x)) |> 
+  st_set_dimensions(3, values = paste0("band", 1:6)) |>
   st_set_dimensions(names = c("x", "y", "band"))
 
 ## -----------------------------------------------------------------------------
@@ -60,14 +59,14 @@ plot(circle, col = NA, border = 'red', add = TRUE, lwd = 2)
 #  image(x3[,,,1])
 
 ## ----eval=ev------------------------------------------------------------------
-system.file("nc/bcsd_obs_1999.nc", package = "stars") %>%
-	read_stars("data/full_data_daily_2013.nc") -> w
+system.file("nc/bcsd_obs_1999.nc", package = "stars") |>
+	read_stars() -> w
 
 ## ----eval=ev------------------------------------------------------------------
 w
 
 ## -----------------------------------------------------------------------------
-system.file("nc/bcsd_obs_1999.nc", package = "stars") %>%
+system.file("nc/bcsd_obs_1999.nc", package = "stars") |>
     read_ncdf()
 
 ## ----eval=ev------------------------------------------------------------------
@@ -90,14 +89,16 @@ file_list = system.file(paste0("netcdf/", x), package = "starsdata")
 ## ----eval=ev------------------------------------------------------------------
 library(dplyr)
 library(abind)
-z <- y %>% select(sst) %>% adrop()
+z <- y |> select(sst) |> adrop()
 
 ## ----eval=ev------------------------------------------------------------------
+# convert POSIXct time to character, to please ggplot's facet_wrap()
+z1 = st_set_dimensions(z, 3, values = as.character(st_get_dimension_values(z, 3)))
 library(ggplot2)
 library(viridis)
 library(ggthemes)
 ggplot() +  
-  geom_stars(data = z[1], alpha = 0.8, downsample = c(10, 10, 1)) + 
+  geom_stars(data = z1[1], alpha = 0.8, downsample = c(10, 10, 1)) + 
   facet_wrap("time") +
   scale_fill_viridis() +
   coord_equal() +
@@ -144,16 +145,32 @@ traffic = array(rpois(prod(n), 10), dim = n) # simulated traffic counts
 (st = st_as_stars(list(traffic = traffic),  dimensions = dims))
 
 ## ----eval=FALSE---------------------------------------------------------------
-#  st %>% as.tbl_cube()
+#  st |> as.tbl_cube()
 
 ## ----eval=FALSE---------------------------------------------------------------
-#  b <- st %>%
-#    as.tbl_cube() %>%
-#    filter(mode == "bike") %>%
-#    group_by(hour) %>%
-#    summarise(traffic = mean(traffic)) %>%
+#  b <- st |>
+#    as.tbl_cube() |>
+#    filter(mode == "bike") |>
+#    group_by(hour) |>
+#    summarise(traffic = mean(traffic)) |>
 #    as.data.frame()
 #  require(ggforce) # for plotting a units variable
 #  ggplot() +
 #    geom_line(data = b, aes(x = hour, y = traffic))
 
+## -----------------------------------------------------------------------------
+s = system.file("tif/lc.tif", package = "stars")
+r = read_stars(s, proxy = FALSE) |> droplevels()
+st_point(c(3190631, 3125)) |> st_sfc(crs = st_crs(r)) |> st_buffer(25000) -> pol1
+st_point(c(3233847, 21027)) |> st_sfc(crs = st_crs(r)) |> st_buffer(10000) -> pol2
+if (isTRUE(dev.capabilities()$rasterImage == "yes")) {
+  plot(r, key.width = lcm(4), reset = FALSE, key.pos = 4)
+  plot(c(pol1, pol2), col = NA, border = c('yellow', 'green'), lwd = 2, add = TRUE)
+}
+
+## -----------------------------------------------------------------------------
+f = function(x) { tb = table(x); names(tb)[which.max(tb)] }
+
+## -----------------------------------------------------------------------------
+aggregate(r, c(pol1, pol2), f) |> st_as_sf()
+
